#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/05/11 15:09


class XlServer:
    """ 厦门理工模式识别与图像理解重点实验室 API服务接口 """

    def __init__(self, host='http://172.16.170.134/', port=5003):
        pass
